﻿using System;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

#pragma warning disable IDE0044

namespace Microsoft.Samples.Eventing.Interop
{
    [Flags]
    internal enum EventEnableProperty : uint
    {
        None = 0,
        Sid = 1,
        TsId = 2,
        StackTrace = 4
    }

    public enum TraceEventLevel : byte
    {
        Always,
        Critical,
        Error,
        Warning,
        Informational,
        Verbose
    }

    [StructLayout(LayoutKind.Explicit)]
    internal struct EVENT_FILTER_DESCRIPTOR
    {
        [FieldOffset(0)]
        public unsafe byte* Ptr;
        [FieldOffset(8)]
        public int Size;
        [FieldOffset(12)]
        public int Type;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct ENABLE_TRACE_PARAMETERS
    {
        public uint Version;
        public uint EnableProperty;
        public uint ControlFlags;
        public Guid SourceId;
        public unsafe EVENT_FILTER_DESCRIPTOR* EnableFilterDesc;
        public int FilterDescCount;
    }

    [StructLayout(LayoutKind.Explicit)]
    internal struct WNodeHeader
    {
        [FieldOffset(0)]
        internal int BufferSize;
        [FieldOffset(4)]
        private uint ProviderId;
        [FieldOffset(8)]
        internal ulong HistoricalContext;
        [FieldOffset(8)]
        private VersionLinkageType VersionLinkage;
        [FieldOffset(16)]
        private IntPtr KernelHandle;
        [FieldOffset(16)]
        private long TimeStamp;
        [FieldOffset(24)]
        internal Guid Guid;
        /// <summary>
        /// Timestamp Type
        /// </summary>
        [FieldOffset(40)]
        internal uint ClientContext;
        [FieldOffset(44)]
        internal uint Flags;

        [StructLayout(LayoutKind.Sequential)]
        private struct VersionLinkageType
        {
            private uint Version;
            private uint Linkage;
        }
    }

#pragma warning disable CA1815
    public struct EventTraceProperitesManaged
    {
        public Guid Guid;
        public ulong SessionID;
        public string SessionName;
        public string LogFileName;

        public uint WnodeFlags;
        internal uint FlushTimer;

        public override string ToString()
        {
            string flags = ((WnodeFlags & EventTraceProperties.WNODE_FLAG_TRACED_GUID) == EventTraceProperties.WNODE_FLAG_TRACED_GUID) ? "WNODE_FLAG_TRACED_GUID" : "";
            return $"{SessionName}(ID = {SessionID:x}, Flags = {flags}, Timer = {FlushTimer})";
        }
    }
#pragma warning restore CA1815

#pragma warning disable CA1815
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    public unsafe struct EventTraceProperties
    {
        static int _recordSize;
        public static int RecordSize
        {
            get
            {
                if (_recordSize == 0)
                {
                    _recordSize = sizeof(EventTraceProperties);
                }

                return _recordSize;
            }
        }

        static int _maxSessionBufferSize;
        public static int MaxSessionBufferSize
        {
            get
            {
                return _maxSessionBufferSize = RecordSize * NativeMethods.MAX_SESSIONS;
            }
        }

        private EventTracePropertiesInternal Internal;

        internal fixed char LogFileNameBuffer[NativeMethods.MAX_LOGFILE_PATH_LEN];
        internal fixed char LoggerNameBuffer[NativeMethods.MAX_SESSION_NAME_LEN];

        internal const uint WNODE_FLAG_TRACED_GUID = 0x00020000;
        internal const int ERROR_WMI_INSTANCE_NOT_FOUND = 4201;

        public unsafe EventTraceProperitesManaged ReadAsManaged()
        {
            EventTraceProperitesManaged item = new EventTraceProperitesManaged()
            {
                WnodeFlags = this.Internal.Wnode.Flags,
                FlushTimer = this.Internal.FlushTimer,
                Guid = this.Internal.Wnode.Guid,
                SessionID = this.Internal.Wnode.HistoricalContext,
            };

            fixed (char* pBuffer = LogFileNameBuffer)
            {
                item.LogFileName = Marshal.PtrToStringUni(new IntPtr(pBuffer));
            }

            fixed (char* pBuffer = LoggerNameBuffer)
            {
                item.SessionName = Marshal.PtrToStringUni(new IntPtr(pBuffer));
            }

            return item;
        }

        internal EventTraceProperties(bool initialize, string sessionName) : this(initialize, sessionName, Guid.Empty, NativeMethods.EVENT_TRACE_REAL_TIME_MODE)
        {
        }

        internal EventTraceProperties(bool initialize, string sessionName, Guid etwProvider) : this(initialize, sessionName, etwProvider, NativeMethods.EVENT_TRACE_REAL_TIME_MODE)
        {
        }

        internal EventTraceProperties(bool initialize, string sessionName, Guid etwProvider, uint logFileMode)
        {
            this.Internal = new EventTracePropertiesInternal();

            if (initialize)
            {
                Initialize(WNODE_FLAG_TRACED_GUID);
            }

            this.Internal.LogFileMode = logFileMode;
            this.Internal.FlushTimer = 1;

            if (etwProvider != Guid.Empty)
            {
                this.Internal.Wnode.Guid = etwProvider;
            }

            if (string.IsNullOrEmpty(sessionName) == false)
            {
                fixed (char* pBuffer = LoggerNameBuffer)
                {
                    Marshal.Copy(sessionName.ToArray(), 0, new IntPtr(pBuffer), sessionName.Length);
                }

                this.Internal.Wnode.ClientContext = NativeMethods.USE_QUERY_PERFORMANCE_COUNTER_TIMESTAMPS;
            }
        }

        public void Initialize(uint flags = 0)
        {
            this.Internal.Wnode.BufferSize = RecordSize;
            System.Diagnostics.Debug.Assert(this.Internal.Wnode.BufferSize == 4216);
            this.Internal.Wnode.Flags = flags;

            this.Internal.LogFileNameOffset = Marshal.SizeOf(typeof(EventTracePropertiesInternal));
            System.Diagnostics.Debug.Assert(this.Internal.LogFileNameOffset == 120);
            this.Internal.LoggerNameOffset = this.Internal.LogFileNameOffset + NativeMethods.MAX_LOGFILE_PATH_LEN * sizeof(char) /*unicode*/;
        }

        internal void SetParameters(uint logFileMode, uint bufferSize, uint minBuffers, uint maxBuffers, uint flushTimerSeconds)
        {
            this.Internal.LogFileMode = logFileMode;

            // Set the buffer size. BufferSize is in KB.
            this.Internal.BufferSize = bufferSize;
            this.Internal.MinimumBuffers = minBuffers;
            this.Internal.MaximumBuffers = maxBuffers;

            // Number of seconds before timer is flushed.
            this.Internal.FlushTimer = flushTimerSeconds;
        }

        internal void SetFlags(uint enableFlags)
        {
            this.Internal.EnableFlags = enableFlags;            
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct EventTracePropertiesInternal
        {
            internal WNodeHeader Wnode;
            internal uint BufferSize;
            internal uint MinimumBuffers;
            internal uint MaximumBuffers;
            private uint MaximumFileSize;
            internal uint LogFileMode;
            internal uint FlushTimer;
            internal uint EnableFlags;
            private int AgeLimit;
            private uint NumberOfBuffers;
            private uint FreeBuffers;
            private uint EventsLost;
            private uint BuffersWritten;
            private uint LogBuffersLost;
            private uint RealTimeBuffersLost;
            private IntPtr LoggerThreadId;
            internal int LogFileNameOffset;
            internal int LoggerNameOffset;
        }
    }
#pragma warning restore CA1815

#pragma warning disable CS0649
    internal struct TRACE_PROVIDER_INFO
    {
        public Guid ProviderGuid;
        public int SchemaSource;
        public int ProviderNameOffset;
    }
#pragma warning restore CS0649

    internal struct PROVIDER_ENUMERATION_INFO
    {
        public int NumberOfProviders;
        public int Padding;
        // TRACE_PROVIDER_INFO TraceProviderInfoArray[ANYSIZE];
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct EtwEventDescriptor
    {
        private ushort Id;
        private byte Version;
        private byte Channel;
        private byte Level;
        internal byte Opcode;
        private ushort Task;
        private ulong Keyword;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct EventHeader
    {
        private ushort Size;
        private ushort HeaderType;
        private ushort Flags;
        private ushort EventProperty;
        public uint ThreadId;
        public uint ProcessId;
        internal Int64 TimeStamp;
        public Guid ProviderId;
        internal EtwEventDescriptor EventDescriptor;
        private ulong ProcessorTime;
        private Guid ActivityId;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct EventRecord
    {
        internal EventHeader EventHeader;
        private EtwBufferContext BufferContext;
        private ushort ExtendedDataCount;
        internal ushort UserDataLength;
        private IntPtr ExtendedData;
        internal IntPtr UserData;
        internal IntPtr UserContext;

        [StructLayout(LayoutKind.Sequential)]
        private struct EtwBufferContext
        {
            private byte ProcessorNumber;
            private byte Alignment;
            private ushort LoggerId;
        }
    }

    internal delegate void EventRecordCallback([In] ref EventRecord eventRecord);

    [StructLayout(LayoutKind.Sequential)]
    internal struct EventTraceHeader
    {
        private ushort Size;
        private ushort FieldTypeFlags;
        private uint Version;
        private uint ThreadId;
        private uint ProcessId;
        private long TimeStamp;
        private Guid Guid;
        private uint KernelTime;
        private uint UserTime;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct EventTrace
    {
        private EventTraceHeader Header;
        private uint InstanceId;
        private uint ParentInstanceId;
        private Guid ParentGuid;
        private IntPtr MofData;
        private uint MofLength;
        private uint ClientContext;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct SystemTime
    {
        private short Year;
        private short Month;
        private short DayOfWeek;
        private short Day;
        private short Hour;
        private short Minute;
        private short Second;
        private short Milliseconds;
    }

    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    internal struct Win32TimeZoneInfo
    {
        private int Bias;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]
        private char[] StandardName;
        private SystemTime StandardDate;
        private int StandardBias;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 32)]
        private char[] DaylightName;
        private SystemTime DaylightDate;
        private int DaylightBias;
    }

    [StructLayout(LayoutKind.Sequential)]
    internal struct TraceLogfileHeader
    {
        private uint BufferSize;
        private uint Version;
        private uint ProviderVersion;
        private uint NumberOfProcessors;
        private long EndTime;
        private uint TimerResolution;
        private uint MaximumFileSize;
        private uint LogFileMode;
        private uint BuffersWritten;
        private Guid LogInstanceGuid;
        private IntPtr LoggerName;
        private IntPtr LogFileName;
        private Win32TimeZoneInfo TimeZone;
        private long BootTime;
        private long PerfFreq;
        private long StartTime;
        private uint ReservedFlags;
        private uint BuffersLost;
    }

    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Unicode)]
    internal struct EventTraceLogfile
    {
        [MarshalAs(UnmanagedType.LPWStr)]
        private string LogFileName;
        [MarshalAs(UnmanagedType.LPWStr)]
        internal string LoggerName;
        private Int64 CurrentTime;
        private uint BuffersRead;
        internal uint ProcessTraceMode;
        private EventTrace CurrentEvent;
        private TraceLogfileHeader LogfileHeader;
        private IntPtr BufferCallback;
        private uint BufferSize;
        private uint Filled;
        private uint EventsLost;
        [MarshalAs(UnmanagedType.FunctionPtr)]
        internal EventRecordCallback EventRecordCallback;
        private uint IsKernelTrace;
        private IntPtr Context;
    }

    unsafe partial class NativeMethods
    {
        public const string KERNEL_LOGGER_NAME = "NT Kernel Logger";
        public static readonly Guid SystemTraceControlGuid = new Guid("{9e814aad-3204-11d2-9a82-006008a86939}");

        public const int MAX_SESSIONS = 64;
        public const int MAX_SESSION_NAME_LEN = 1024;
        public const int MAX_LOGFILE_PATH_LEN = 1024;

        public const uint EVENT_TRACE_REAL_TIME_MODE = 0x00000100;
        public const uint EVENT_TRACE_NO_PER_PROCESSOR_BUFFERING = 0x10000000;

        public const uint PROCESS_TRACE_MODE_REAL_TIME = 0x00000100; // == EVENT_TRACE_REAL_TIME_MODE
        public const uint PROCESS_TRACE_MODE_RAW_TIMESTAMP = 0x00001000;
        public const uint PROCESS_TRACE_MODE_EVENT_RECORD = 0x10000000;

        public const uint EVENT_TRACE_CONTROL_QUERY = 0;
        public const uint EVENT_TRACE_CONTROL_STOP = 1;
        public const uint EVENT_TRACE_CONTROL_UPDATE = 2;

        public const uint USE_QUERY_PERFORMANCE_COUNTER_TIMESTAMPS = 1;

        public const uint EVENT_TRACE_FLAG_PROCESS = 0x00000001;

        public const uint EVENT_CONTROL_CODE_ENABLE_PROVIDER = 1;
        public const uint EVENT_CONTROL_CODE_DISABLE_PROVIDER = 0;

        [DllImport("tdh.dll")]
        internal static extern int TdhEnumerateProviders(PROVIDER_ENUMERATION_INFO* pBuffer, ref int pBufferSize);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern int QueryAllTraces(IntPtr pPropertyArray, uint propertyArrayCount, ref uint loggerCount);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern int QueryTrace(
            [In] ulong traceHandle,
            [In, MarshalAs(UnmanagedType.LPWStr)] string instanceName,
            [In, Out] ref EventTraceProperties eventTraceProperties);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern int ControlTrace(
            [In] ulong traceHandle,
            [In, MarshalAs(UnmanagedType.LPWStr)] string instanceName,
            [In, Out] ref EventTraceProperties eventTraceProperties,
            [In] uint controlCode);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern unsafe int StartTrace(out ulong sessionHandle, [In] string sessionName, ref EventTraceProperties properties);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern ulong /*session handle*/ OpenTrace([In, Out] ref EventTraceLogfile logfile);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern int ProcessTrace([In] ulong[] handleArray, [In] uint handleCount, [In] IntPtr StartTime, [In] IntPtr EndTime);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern int CloseTrace([In] ulong traceHandle);

        [DllImport("advapi32.dll", CharSet = CharSet.Unicode)]
        internal static extern int EnableTraceEx2([In] ulong TraceHandle, [In] ref Guid ProviderId, [In] uint ControlCode, [In] byte Level, [In] ulong MatchAnyKeyword, [In] ulong MatchAllKeyword, [In] int Timeout, [In] ref ENABLE_TRACE_PARAMETERS EnableParameters);
    }
}
